"""A flexible caching decorator.

This module provides a cache decorator `cache` that you can use to
cache results of your functions or methods.  Let's say we have a class
with an expensive method `pow` that we want to cache:

  >>> class MyClass:
  ...     def pow(self, first, second):
  ...         print 'Someone or something called me'
  ...         return first ** second

Okay, we know that if the `first` and `second` arguments are the same,
the result is going to be the same, always.  We'll use a cache key
calculator to tell the `cache` decorator about this assertion.  What's
this cache key calculator?  It's a function that takes the original
function plus the same arguments as the original function that we're
caching:

  >>> def cache_key(method, self, first, second):
  ...     return hash((first, second))

The cache decorator is really simple to use.  Let's define our first
class again, this time with a cached `pow` method:

  >>> class MyClass:
  ...     @cache(cache_key)
  ...     def pow(self, first, second):
  ...         print 'Someone or something called me'
  ...         return first ** second

The results:

  >>> obj = MyClass()
  >>> obj.pow(3, 2)
  Someone or something called me
  9
  >>> obj.pow(3, 2)
  9

Did you see that?  The method was called only once.

Now to where this cache is stored: That's actually variable.  The
cache decorator takes an optional second argument with which you can
define the where the cache is stored to.

By default, the cache stores its values on the first argument to the
function.  For our method, this is self, which is perfectly fine.  For
normal functions, the first argument is maybe not the best place to
store the cache.

The default cache container function stores a dictionary on the
instance as a *volatile* attribute.  That is, it's prefixed with
``_v_``.  In Zope, this means that the cache is not persisted.

  >>> ATTR
  '_v_memoize_cache'
  >>> cache_container = getattr(obj, ATTR)

This cache container maps our key, including the function's dotted
name, to the return value.

  >>> cache_container # doctest: +ELLIPSIS
  {'plone.memoize.volatile.pow:...': 9}
  >>> len(cache_container)
  1
  >>> k = 'plone.memoize.volatile.pow:%s' % cache_key(MyClass.pow, None, 3, 2)
  >>> cache_container[k]
  9

Okay, on to storing the cache somewhere else.  The function we'll have
to provide is really similar to the cache key function we defined
earlier.

Like the cache key function, the storage function takes the same
amount of arguments as the original cached function.  We'll use a
global for caching this time:

  >>> my_cache = {}
  >>> def cache_storage(fun, *args, **kwargs):
  ...     return my_cache

This time, instead of caching a method, we'll cache a normal function.
For this, we'll need to change our cache key function to take the
correct number of arguments:

  >>> def cache_key(fun, first, second):
  ...     return hash((first, second))    

Note how we provide both the cache key generator and the cache storage
as arguments to the `cache` decorator:

  >>> @cache(cache_key, cache_storage)
  ... def pow(first, second):
  ...     print 'Someone or something called me'
  ...     return first ** second

Let's try it out:

  >>> pow(3, 2)
  Someone or something called me
  9
  >>> pow(3, 2)
  9
  >>> pow(3, 2)
  9
  >>> pow(3, 3)
  Someone or something called me
  27
  >>> pow(3, 3)
  27
  >>> my_cache.clear()

It works!

A cache key generator may also raise DontCache to indicate that no
caching should be applied:

  >>> def cache_key(fun, first, second):
  ...     if first == second:
  ...         raise DontCache
  ...     else:
  ...         return hash((first, second))    
  >>> @cache(cache_key, cache_storage)
  ... def pow(first, second):
  ...     print 'Someone or something called me'
  ...     return first ** second

  >>> pow(3, 2)
  Someone or something called me
  9
  >>> pow(3, 2)
  9
  >>> pow(3, 3)
  Someone or something called me
  27
  >>> pow(3, 3)
  Someone or something called me
  27

Caveats
-------

Be careful when you have multiple methods with the same name in a
single module:

  >>> def cache_key(fun, instance, *args):
  ...     return args
  >>> cache_container = {}
  >>> class A:
  ...     @cache(cache_key, lambda *args: cache_container)
  ...     def somemet(self, one, two):
  ...         return one + two
  >>> class B:
  ...     @cache(cache_key, lambda *args: cache_container)
  ...     def somemet(self, one, two):
  ...         return one - two
  >>> a = A()
  >>> a.somemet(1, 2)
  3
  >>> cache_container
  {'plone.memoize.volatile.somemet:(1, 2)': 3}

The following call should really return -1, but since the default
cache key isn't clever enough to include the function's name, it'll
return 3:

  >>> B().somemet(1, 2)
  3
  >>> len(cache_container)
  1
  >>> cache_container.clear()

Ouch!  The fix for this is to e.g. include your class' name in the key
when you create it:

  >>> def cache_key(fun, instance, *args):
  ...     return (instance.__class__,) + args
  >>> class A:
  ...     @cache(cache_key, lambda *args: cache_container)
  ...     def somemet(self, one, two):
  ...         return one + two
  >>> class B:
  ...     @cache(cache_key, lambda *args: cache_container)
  ...     def somemet(self, one, two):
  ...         return one - two
  >>> a = A()
  >>> a.somemet(1, 2)
  3
  >>> B().somemet(1, 2)
  -1
  >>> len(cache_container)
  2
"""

import time

class CleanupDict(dict):
    """A dict that automatically cleans up items that haven't been
    accessed in a given timespan on *set*.

    This implementation is a bit naive, since it's not associated with
    any policy that the user can configure, and it doesn't provide
    statistics like RAMCache, but at least it helps make sure our
    volatile attribute doesn't grow stale entries indefinitely.

      >>> d = CleanupDict()
      >>> d['spam'] = 'bar'
      >>> d['spam']
      'bar'

      >>> d = CleanupDict(0)
      >>> d['spam'] = 'bar'
      >>> d['spam'] # doctest: +ELLIPSIS
      Traceback (most recent call last):
      ...
      KeyError: 'spam'
    """
    cleanup_period = 60 * 60 * 24 * 3 # 3 days

    def __init__(self, cleanup_period=None):
        super(CleanupDict, self).__init__()
        self._last_access = {}
        if cleanup_period is not None:
            self.cleanup_period = cleanup_period

    def __getitem__(self, key):
        value = super(CleanupDict, self).__getitem__(key)
        self._last_access[key] = time.time()
        return value

    def __setitem__(self, key, value):
        super(CleanupDict, self).__setitem__(key, value)
        self._last_access[key] = time.time()
        self._cleanup()

    def _cleanup(self):
        now = time.time()
        okay = now - self.cleanup_period
        for key, timestamp in self._last_access.items():
            if timestamp < okay:
                del self._last_access[key]
                super(CleanupDict, self).__delitem__(key)
    
ATTR = '_v_memoize_cache'
CONTAINER_FACTORY = CleanupDict
_marker = object()

class DontCache(Exception):
    pass

def store_on_self(method, obj, *args, **kwargs):
    return obj.__dict__.setdefault(ATTR, CONTAINER_FACTORY())

def store_on_context(method, obj, *args, **kwargs):
    return obj.context.__dict__.setdefault(ATTR, CONTAINER_FACTORY())

def cache(get_key, get_cache=store_on_self):
    def decorator(fun):
        def replacement(*args, **kwargs):
            try:
                key = get_key(fun, *args, **kwargs)
            except DontCache:
                return fun(*args, **kwargs)
            key = '%s.%s:%s' % (fun.__module__, fun.__name__, key)
            cache = get_cache(fun, *args, **kwargs)
            cached_value = cache.get(key, _marker)
            if cached_value is _marker:
                cached_value = cache[key] = fun(*args, **kwargs)
            return cached_value
        return replacement
    return decorator



"""
Memo decorators for globals - memoized values survive for as long as the
process lives.

Stores values in a module-level variable.
"""

from plone.memoize import volatile

_memos = {}

def memoize(fun):
    def get_key(fun, *args, **kwargs):
        return hash((args, frozenset(kwargs.items()),))
    def get_cache(fun, *args, **kwargs):
        return _memos
    return volatile.cache(get_key, get_cache)(fun)

__all__ = (memoize,)



"""A cache decorator that uses RAMCache by default.

See README.txt and the `volatile` module for more details.

  >>> def cache_key(fun, first, second):
  ...     return hash((first, second))
  >>> @cache(cache_key)
  ... def pow(first, second):
  ...     print 'Someone or something called me'
  ...     return first ** second

  >>> pow(3, 2)
  Someone or something called me
  9
  >>> pow(3, 2)
  9

Let's cache another function:

  >>> @cache(cache_key)
  ... def add(first, second):
  ...     print 'Someone or something called me'
  ...     return first + second

  >>> add(3, 2)
  Someone or something called me
  5
  >>> add(3, 2)
  5

Now invalidate the cache for the `pow` function:

  >>> pow(3, 2)
  9
  >>> global_cache.invalidate('plone.memoize.ram.pow')
  >>> pow(3, 2)
  Someone or something called me
  9

Make sure that we only invalidated the cache for the `pow` function:

  >>> add(3, 2)
  5

  >>> global_cache.invalidateAll()

You can register an ICacheChooser utility to override the cache used
based on the function that is cached.  To do this, we'll first
unregister the already registered global `choose_cache` function:

  >>> sm = component.getGlobalSiteManager()
  >>> sm.unregisterUtility(choose_cache)
  True

This customized cache chooser will use the `my_cache` for the `pow`
function, and use the `global_cache` for all other functions:

  >>> my_cache = ram.RAMCache()
  >>> def my_choose_cache(fun_name):
  ...     if fun_name.endswith('.pow'):
  ...         return RAMCacheAdapter(my_cache)
  ...     else:
  ...         return RAMCacheAdapter(global_cache)
  >>> interface.directlyProvides(my_choose_cache, ICacheChooser)
  >>> sm.registerUtility(my_choose_cache)

Both caches are empty at this point:

  >>> len(global_cache.getStatistics())
  0
  >>> len(my_cache.getStatistics())
  0

Let's fill them:

  >>> pow(3, 2)
  Someone or something called me
  9
  >>> pow(3, 2)
  9
  >>> len(global_cache.getStatistics())
  0
  >>> len(my_cache.getStatistics())
  1

  >>> add(3, 2)
  Someone or something called me
  5
  >>> add(3, 2)
  5
  >>> len(global_cache.getStatistics())
  1
  >>> len(my_cache.getStatistics())
  1
"""

import md5
import cPickle

from zope import interface
from zope import component
from zope.app.cache.interfaces.ram import IRAMCache
from zope.app.cache import ram

from plone.memoize.interfaces import ICacheChooser
from plone.memoize import volatile

global_cache = ram.RAMCache()
global_cache.update(maxAge=86400)

DontCache = volatile.DontCache
MARKER = object()

class AbstractDict:
    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

class MemcacheAdapter(AbstractDict):
    def __init__(self, client, globalkey=''):
        self.client = client
        self.globalkey = globalkey and '%s:' % globalkey

    def _make_key(self, source):
        return md5.new(source).hexdigest()

    def __getitem__(self, key):
        cached_value = self.client.get(self.globalkey + self._make_key(key))
        if cached_value is None:
            raise KeyError(key)
        else:
            return cPickle.loads(cached_value)

    def __setitem__(self, key, value):
        cached_value = cPickle.dumps(value)
        self.client.set(self.globalkey + self._make_key(key), cached_value)

class RAMCacheAdapter(AbstractDict):
    def __init__(self, ramcache, globalkey=''):
        self.ramcache = ramcache
        self.globalkey = globalkey

    def _make_key(self, source):
        return md5.new(source).digest()

    def __getitem__(self, key):
        value = self.ramcache.query(self.globalkey,
                                    dict(key=self._make_key(key)),
                                    MARKER)
        if value is MARKER:
            raise KeyError(key)
        else:
            return value

    def __setitem__(self, key, value):
        self.ramcache.set(value,
                          self.globalkey,
                          dict(key=self._make_key(key)))

def choose_cache(fun_name):
    return RAMCacheAdapter(component.queryUtility(IRAMCache),
                           globalkey=fun_name)
interface.directlyProvides(choose_cache, ICacheChooser)

def store_in_cache(fun, *args, **kwargs):
    key = '%s.%s' % (fun.__module__, fun.__name__)
    cache = None
    cache_chooser = component.queryUtility(ICacheChooser)
    if cache_chooser is not None:
        cache = cache_chooser(key)
    if cache is None:
        return {}
    else:
        return cache

def cache(get_key):
    return volatile.cache(get_key, get_cache=store_in_cache)



"""
In order to use Peter Bengtsson's slimmer library available from
http://www.issuetrackerproduct.com/Download#slimmer
you need to register the XHTMLSlimmer utility like this:

  <utility component="plone.memoize.compress.xhtmlslimmer"
           provides="plone.memoize.interfaces.IXHTMLCompressor" />

You can register other XHTML-whitespace removal libraries as well.

Per default there is no compressor available and we get the same string back:

   >>> html_string = u"<html><body><SPAN>Hello.</SPAN></body><html>"
   >>> xhtml_compress(html_string) is html_string
   True

Make a stupid lowercasing compressor. This is not safe as it would lowercase
all text outside of tags as well:

   >>> class LowerCaser(object):
   ...     implements(IXHTMLCompressor)
   ...
   ...     def compress(self, string):
   ...         return string.lower()

   >>> lower = LowerCaser()

Register our new compressor:

   >>> from zope.component import getSiteManager
   >>> sm = getSiteManager()
   >>> sm.registerUtility(lower)

   >>> xhtml_compress(html_string)
   u'<html><body><span>hello.</span></body><html>'
"""

from zope.component import queryUtility
from zope.interface import implements

from plone.memoize.interfaces import IXHTMLCompressor

SLIMMER = True
try:
    from slimmer import xhtml_slimmer
except ImportError:
    SLIMMER = False


def xhtml_compress(string):
    util = queryUtility(IXHTMLCompressor)
    if util is not None:
        return util.compress(string)
    return string


class XHTMLSlimmer(object):

    implements(IXHTMLCompressor)

    def compress(self, string):
        if SLIMMER:
            return xhtml_slimmer(string)
        return string

xhtmlslimmer = XHTMLSlimmer()



"""
Memoize decorator for methods.

Stores values in an annotation of the request.
"""

import inspect

from zope.annotation.interfaces import IAnnotations

from plone.memoize import volatile

_marker = object()
class RequestMemo(object):
    
    key = 'plone.memoize_request'

    def __init__(self, arg=0):
        self.arg = arg

    def __call__(self, func):
        def memogetter(*args, **kwargs):
            request = None
            if isinstance(self.arg, int):
                request = args[self.arg]
            else:
                request = kwargs[self.arg]

            annotations = IAnnotations(request)
            cache = annotations.get(self.key, _marker)
        
            if cache is _marker:
                cache = annotations[self.key] = dict()
        
            key = hash((func.__module__, func.__name__, 
                        args, frozenset(kwargs.items())),)
            value = cache.get(key, _marker)
            if value is _marker:
                value = cache[key] = func(*args, **kwargs)
            return value
        return memogetter

def store_in_annotation_of(expr):
    def _store_in_annotation(fun, *args, **kwargs):
        # Use expr to find out the name of the request variable
        vars = {}
        for index, name in enumerate(inspect.getargspec(fun)[0]):
            if name in kwargs:
                vars[name] = kwargs[name]
            else:
                vars[name] = args[index]
        request = eval(expr, {}, vars)
        return IAnnotations(request)
    return _store_in_annotation

def cache(get_key, get_request='request'):
    r"""
    This is a hypothetical function `increment` that'll store the
    cache value on `a.request`, where a is the only argument to the
    function:
    
      >>> def increment(a):
      ...     print 'Someone or something called me'
      ...     return a + 1

    Now we need to define this `a`.  For this, we'll inherit from
    `int` and add a `request` class variable.  Note that we also make
    our fake request `IAttributeAnnotatable`, because that's how the
    cache values are stored on the request:

      >>> from zope.publisher.browser import TestRequest
      >>> class A(int):
      ...     request = TestRequest()
      >>> from zope.interface import directlyProvides
      >>> from zope.annotation.interfaces import IAttributeAnnotatable
      >>> directlyProvides(A.request, IAttributeAnnotatable)

    In addition to this request, we'll also need to set up a cache key
    generator.  We'll use the integer value of the only argument for
    that:

      >>> get_key = lambda fun, a, *args: a

    Let's decorate our `increment` function now with the `cache`
    decorator.  We'll tell the decorator to use `args_hash` for
    generating the key. `get_request` will tell the decorator how to
    actually find the `request` in the variable scope of the function
    itself:
    
      >>> cached_increment = \
      ...     cache(get_key=get_key, get_request='a.request')(increment)

      >>> cached_increment(A(1))
      Someone or something called me
      2
      >>> cached_increment(A(1))
      2
      >>> IAnnotations(A.request)
      {'plone.memoize.request.increment:1': 2}

    If `request` is already part of the function's argument list, we
    don't need to specify any expression:

      >>> @cache(get_key=get_key)
      ... def increment_plus(a, request):
      ...     print 'Someone or something called me'
      ...     return a + 1

      >>> increment_plus(42, A.request)
      Someone or something called me
      43
      >>> increment_plus(42, A.request)
      43
      >>> IAnnotations(A.request)['plone.memoize.request.increment_plus:42']
      43
    """

    return volatile.cache(get_key,
                          get_cache=store_in_annotation_of(get_request))

memoize_diy_request = RequestMemo

__all__ = (memoize_diy_request, store_in_annotation_of, cache)



from setuptools import setup, find_packages
import sys, os

version = '1.0.1.1'

setup(name='plone.memoize',
      version=version,
      description="Decorator to memoize properties bound to a request",
      long_description="""\
plone.memoize provides memo decorators for Zope 3 views and regular 
classes. For views, a memo will last as long as the request, even
if the view is looked up several times. This requires the request
to be marked as annotatable.
""",
      classifiers=[],
      keywords='plone memoize decorator',
      author='Martin Aspeli and David "Whit" Morriss',
      author_email='optilude@gmx.net',
      url='http://svn.plone.org/svn/plone/plone.memoize',
      license='LGPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['plone'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'setuptools',
      ],
      )



"""
Memoize decorator for views.

Stores values in an annotation of the request. See view.txt.
"""

from zope.annotation.interfaces import IAnnotations, IAttributeAnnotatable
from zope.interface import alsoProvides

_marker = object()
class ViewMemo(object):
    
    key = 'plone.memoize'
    
    def memoize(self, func):
        def memogetter(*args, **kwargs):
            instance = args[0]
            
            context = getattr(instance, 'context', None)            
            request = getattr(instance, 'request', None)
            
            # XXX: A Zope 2 workaround/optimisation (will degrade gracefully)
            if request is None:
                request = getattr(context, 'REQUEST', None)
            
            annotations = IAnnotations(request)
            cache = annotations.get(self.key, _marker)
            
            if cache is _marker:
                cache = annotations[self.key] = dict()
                
            # XXX: Not the most elegant thing in the world; in a Zope 2
            # context, the physical path is a better key, since the id could
            # change if the object is invalidated from the ZODB cache
            
            try:
                context_id = context.getPhysicalPath()
            except AttributeError:
                context_id = id(context)
            
            # Note: we don't use args[0] in the cache key, since args[0] == 
            # instance and the whole point is that we can cache different
            # requests
            
            key = hash((context_id, instance.__class__.__name__, func.__name__, 
                        args[1:], frozenset(kwargs.items())),)
            value = cache.get(key, _marker)
            if value is _marker:
                value = cache[key] = func(*args, **kwargs)
            return value
        return memogetter

    def memoize_contextless(self, func):
        def memogetter(*args, **kwargs):
            instance = args[0]
            request = getattr(instance, 'request', None)
            
            annotations = IAnnotations(request)
            cache = annotations.get(self.key, _marker)
            
            if cache is _marker:
                cache = annotations[self.key] = dict()
            
            key = hash((instance.__class__.__name__, func.__name__, 
                        args[1:], frozenset(kwargs.items())),)
            value = cache.get(key, _marker)
            if value is _marker:
                value = cache[key] = func(*args, **kwargs)
            return value
        return memogetter

_m = ViewMemo()
memoize = _m.memoize
memoize_contextless = _m.memoize_contextless

__all__ = (memoize, memoize_contextless)


import os, sys, unittest

from zope.testing import doctest

from zope.app.testing.placelesssetup import setUp, tearDown
from zope.configuration.xmlconfig import XMLConfig

optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS

import zope.app.component
import plone.memoize

def configurationSetUp(test):
    setUp()    
    
    XMLConfig('meta.zcml', zope.app.component)()
    XMLConfig('configure.zcml', plone.memoize)()

def configurationTearDown(test):
    tearDown()

def test_suite():
    try:
        from zope.publisher.interfaces.browser import IBrowserView, IBrowserRequest
    except ImportError:
        from zope.app.publisher.interfaces.browser import IBrowserView
        from zope.publisher.interfaces.browser import IBrowserRequest
    from zope.component import adapts
    from zope.component import provideAdapter
    from zope.interface import implements, Interface

    return unittest.TestSuite((
        doctest.DocTestSuite('plone.memoize.compress',
                             setUp=configurationSetUp,
                             tearDown=configurationTearDown),
        doctest.DocFileSuite('instance.txt', 
                             package="plone.memoize",
                             setUp=configurationSetUp,
                             tearDown=configurationTearDown,
                             optionflags=optionflags,
                             globs=locals()),
        doctest.DocFileSuite('view.txt', 
                             package="plone.memoize",
                             setUp=configurationSetUp,
                             tearDown=configurationTearDown,
                             optionflags=optionflags,
                             globs=locals()),
        doctest.DocFileSuite('forever.txt', 
                             package="plone.memoize",
                             setUp=configurationSetUp,
                             tearDown=configurationTearDown,
                             optionflags=optionflags,
                             globs=locals()),
        doctest.DocFileSuite('README.txt'),
        doctest.DocTestSuite('plone.memoize.request',
                             setUp=configurationSetUp,
                             tearDown=configurationTearDown),
        doctest.DocTestSuite('plone.memoize.volatile'),
        doctest.DocTestSuite('plone.memoize.ram',
                             setUp=configurationSetUp,
                             tearDown=configurationTearDown),
        ))

if __name__=="__main__":
    import unittest
    unittest.TextTestRunner().run(test_suite())
